/*
 * QEMU NEC PC-98x1 System Emulator
 *
 * Copyright (c) 2009 TAKEDA, toshiya
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include "hw.h"
#include "pc.h"
#include "fdc.h"
#include "pci.h"
#include "block.h"
#include "sysemu.h"
#include "audio/audio.h"
#include "net.h"
#include "smbus.h"
#include "boards.h"
#include "monitor.h"

#define BIOS_FILENAME "pc98bios.bin"
#define BIOS_FILESIZE 0x18000
#define ITF_FILENAME "pc98itf.bin"
#define ITF_FILESIZE 0x8000
#define IDE_FILENAME "pc98ide.bin"
#define IDE_FILESIZE 0x2000

#define MAX_IDE_BUS 2

static fdctrl_t *floppy_controller;
static PITState *pit;
static IOAPICState *ioapic;
static PCIDevice *i440fx_state;

static ram_addr_t ram_addr[512];
static ram_addr_t bios_addr;
static ram_addr_t itf_addr;
static ram_addr_t ide_addr;

static uint8_t bios_ram_selected;
static uint8_t itf_selected;
static uint8_t ram_window_map;

/* MSDOS compatibility mode FPU exception support */
static qemu_irq ferr_irq;
/* XXX: add IGNNE support */
void cpu_set_ferr(CPUX86State *s)
{
    qemu_irq_raise(ferr_irq);
}

/* TSC handling */
uint64_t cpu_get_tsc(CPUX86State *env)
{
    /* Note: when using kqemu, it is more logical to return the host TSC
       because kqemu does not trap the RDTSC instruction for
       performance reasons */
#ifdef CONFIG_KQEMU
    if (env->kqemu_enabled) {
        return cpu_get_real_ticks();
    } else
#endif
    {
        return cpu_get_ticks();
    }
}

/* SMM support */
void cpu_smm_update(CPUState *env)
{
    if (i440fx_state && env == first_cpu)
        i440fx_set_smm(i440fx_state, (env->hflags >> HF_SMM_SHIFT) & 1);
}


/* IRQ handling */
int cpu_get_pic_interrupt(CPUState *env)
{
    int intno;

    intno = apic_get_interrupt(env);
    if (intno >= 0) {
        /* set irq request if a PIC irq is still pending */
        /* XXX: improve that */
        pic_update_irq(isa_pic);
        return intno;
    }
    /* read the irq from the PIC */
    if (!apic_accept_pic_intr(env))
        return -1;

    intno = pic_read_irq(isa_pic);
    return intno;
}

static void pic_irq_request(void *opaque, int irq, int level)
{
    CPUState *env = first_cpu;

    if (env->apic_state) {
        while (env) {
            if (apic_accept_pic_intr(env))
                apic_deliver_pic_intr(env, level);
            env = env->next_cpu;
        }
    } else {
        if (level)
            cpu_interrupt(env, CPU_INTERRUPT_HARD);
        else
            cpu_reset_interrupt(env, CPU_INTERRUPT_HARD);
    }
}

void ioport_set_a20(int enable)
{
    /* XXX: send to all CPUs ? */
    cpu_x86_set_a20(first_cpu, enable);
}

int ioport_get_a20(void)
{
    return ((first_cpu->a20_mask >> 20) & 1);
}

static void main_cpu_reset(void *opaque)
{
    CPUState *env = opaque;

    if (!itf_selected) {
        cpu_register_physical_memory(0x000f8000, 0x8000, itf_addr | IO_MEM_ROM);
        cpu_register_physical_memory(0xffff8000, 0x8000, itf_addr | IO_MEM_ROM);
        itf_selected = 1;
    }
    cpu_reset(env);
}

/*
#define NE2000_NB_MAX 6

static int ne2000_io[NE2000_NB_MAX] = { 0x300, 0x320, 0x340, 0x360, 0x280, 0x380 };
static int ne2000_irq[NE2000_NB_MAX] = { 9, 10, 11, 3, 4, 5 };

static int serial_io[MAX_SERIAL_PORTS] = { 0x3f8, 0x2f8, 0x3e8, 0x2e8 };
static int serial_irq[MAX_SERIAL_PORTS] = { 4, 3, 4, 3 };

static int parallel_io[MAX_PARALLEL_PORTS] = { 0x378, 0x278, 0x3bc };
static int parallel_irq[MAX_PARALLEL_PORTS] = { 7, 7, 7 };

#ifdef HAS_AUDIO
static void audio_init (PCIBus *pci_bus, qemu_irq *pic)
{
    struct soundhw *c;
    int audio_enabled = 0;

    for (c = soundhw; !audio_enabled && c->name; ++c) {
        audio_enabled = c->enabled;
    }

    if (audio_enabled) {
        AudioState *s;

        s = AUD_init ();
        if (s) {
            for (c = soundhw; c->name; ++c) {
                if (c->enabled) {
                    if (c->isa) {
                        c->init.init_isa (s, pic);
                    }
                    else {
                        if (pci_bus) {
                            c->init.init_pci (pci_bus, s);
                        }
                    }
                }
            }
        }
    }
}
#endif

static void pc_init_ne2k_isa(NICInfo *nd, qemu_irq *pic)
{
    static int nb_ne2k = 0;

    if (nb_ne2k == NE2000_NB_MAX)
        return;
    isa_ne2000_init(ne2000_io[nb_ne2k], pic[ne2000_irq[nb_ne2k]], nd);
    nb_ne2k++;
}
*/

static void ioportF0_write(void *opaque, uint32_t addr, uint32_t data)
{
/*
    if (!itf_selected) {
        cpu_register_physical_memory(0x000f8000, 0x8000, itf_addr | IO_MEM_ROM);
        cpu_register_physical_memory(0xffff8000, 0x8000, itf_addr | IO_MEM_ROM);
        itf_selected = 1;
    }
*/
    qemu_cpu_reset_request();
}

static void ioportF2_write(void *opaque, uint32_t addr, uint32_t data)
{
    ioport_set_a20(1);
}

static uint32_t ioportF2_read(void *opaque, uint32_t addr)
{
    return (ioport_get_a20() ^ 1) | 0x2e;
}

static void ioportF6_write(void *opaque, uint32_t addr, uint32_t data)
{
    switch (data) {
    case 0x02:
        ioport_set_a20(1);
        break;
    case 0x03:
        ioport_set_a20(0);
        break;
    }
}

static uint32_t ioportF6_read(void *opaque, uint32_t addr)
{
    return (ioport_get_a20() ^ 1) | 0x5e;
}

static void ioport43D_write(void *opaque, uint32_t addr, uint32_t data)
{
    switch (data) {
    case 0x00:
    case 0x10:
        if (!itf_selected) {
            cpu_register_physical_memory(0x000f8000, 0x8000, itf_addr | IO_MEM_ROM);
            cpu_register_physical_memory(0xffff8000, 0x8000, itf_addr | IO_MEM_ROM);
            itf_selected = 1;
        }
        break;
    case 0x02:
    case 0x12:
        if (itf_selected) {
            if (bios_ram_selected) {
                cpu_register_physical_memory(0x000e8000, 0x8000, ram_addr[0xe8000 >> 15]);
                cpu_register_physical_memory(0x000f0000, 0x8000, ram_addr[0xf0000 >> 15]);
                cpu_register_physical_memory(0x000f8000, 0x8000, ram_addr[0xf8000 >> 15]);
                cpu_register_physical_memory(0xfffe8000, 0x8000, ram_addr[0xe8000 >> 15]);
                cpu_register_physical_memory(0xffff0000, 0x8000, ram_addr[0xf0000 >> 15]);
                cpu_register_physical_memory(0xffff8000, 0x8000, ram_addr[0xf8000 >> 15]);
            } else {
                cpu_register_physical_memory(0x000e8000, 0x18000, bios_addr | IO_MEM_ROM);
                cpu_register_physical_memory(0xfffe8000, 0x18000, bios_addr | IO_MEM_ROM);
            }
            itf_selected = 0;
        }
        break;
    }
}

static void ioport461_write(void *opaque, uint32_t addr, uint32_t data)
{
    if ((ram_window_map & 0xfe) != (data & 0xfe)) {
        uint32_t bank = (data & 0xfe) << 1;
        cpu_register_physical_memory(0x00080000, 0x8000, ram_addr[bank + 0]);
        cpu_register_physical_memory(0x00088000, 0x8000, ram_addr[bank + 1]);
        cpu_register_physical_memory(0x00090000, 0x8000, ram_addr[bank + 2]);
        cpu_register_physical_memory(0x00098000, 0x8000, ram_addr[bank + 3]);
        cpu_register_physical_memory(0xfff80000, 0x8000, ram_addr[bank + 0]);
        cpu_register_physical_memory(0xfff88000, 0x8000, ram_addr[bank + 1]);
        cpu_register_physical_memory(0xfff90000, 0x8000, ram_addr[bank + 2]);
        cpu_register_physical_memory(0xfff98000, 0x8000, ram_addr[bank + 3]);
        ram_window_map = data;
    }
}

static uint32_t ioport461_read(void *opaque, uint32_t addr)
{
    return ram_window_map;
}

static uint32_t ioport534_read(void *opaque, uint32_t addr)
{
    return 0xec;
}

static void ioport53D_write(void *opaque, uint32_t addr, uint32_t data)
{
    if (!bios_ram_selected && (data & 0x02)) {
        cpu_register_physical_memory(0x000c0000, 0x8000, ram_addr[0xc0000 >> 15]);
        cpu_register_physical_memory(0x000c8000, 0x8000, ram_addr[0xc8000 >> 15]);
        cpu_register_physical_memory(0x000d0000, 0x8000, ram_addr[0xd0000 >> 15]);
        cpu_register_physical_memory(0x000d8000, 0x8000, ram_addr[0xd8000 >> 15]);
        cpu_register_physical_memory(0xfffc0000, 0x8000, ram_addr[0xc0000 >> 15]);
        cpu_register_physical_memory(0xfffc8000, 0x8000, ram_addr[0xc8000 >> 15]);
        cpu_register_physical_memory(0xfffd0000, 0x8000, ram_addr[0xd0000 >> 15]);
        cpu_register_physical_memory(0xfffd8000, 0x8000, ram_addr[0xd8000 >> 15]);
        if (!itf_selected) {
            cpu_register_physical_memory(0x000e8000, 0x8000, ram_addr[0xe8000 >> 15]);
            cpu_register_physical_memory(0x000f0000, 0x8000, ram_addr[0xf0000 >> 15]);
            cpu_register_physical_memory(0x000f8000, 0x8000, ram_addr[0xf8000 >> 15]);
            cpu_register_physical_memory(0xfffe8000, 0x8000, ram_addr[0xe8000 >> 15]);
            cpu_register_physical_memory(0xffff0000, 0x8000, ram_addr[0xf0000 >> 15]);
            cpu_register_physical_memory(0xffff8000, 0x8000, ram_addr[0xf8000 >> 15]);
        }
        bios_ram_selected = 1;
    } else if (bios_ram_selected && !(data & 0x02)) {
        cpu_register_physical_memory(0x000d8000, 0x2000, ide_addr | IO_MEM_ROM);
        cpu_register_physical_memory(0xfffd8000, 0x2000, ide_addr | IO_MEM_ROM);
        if (!itf_selected) {
                cpu_register_physical_memory(0x000e8000, 0x18000, bios_addr | IO_MEM_ROM);
                cpu_register_physical_memory(0xfffe8000, 0x18000, bios_addr | IO_MEM_ROM);
        }
        bios_ram_selected = 0;
    }
}

static uint32_t ioport63D_read(void *opaque, uint32_t addr)
{
    return 0x04;
}

static uint32_t ioport9894_read(void *opaque, uint32_t addr)
{
    return 0x90;
}

/* PC-98 hardware initialisation */
static void pc98_init1(ram_addr_t ram_size, int pci_enabled)
{
    char buf[1024];
    int ret, i;
    uint32_t addr;
    PCIBus *pci_bus;
    int piix3_devfn = -1;
    CPUState *env = NULL;
    qemu_irq *cpu_irq;
    qemu_irq *i8259;
    int index;
    BlockDriverState *hd[MAX_IDE_BUS * MAX_IDE_DEVS];
    BlockDriverState *fd[MAX_FD];

    /* init CPUs */
    for(i = 0; i < smp_cpus; i++) {
        env = cpu_init("qemu32");
        if (!env) {
            fprintf(stderr, "Unable to find x86 CPU definition\n");
            exit(1);
        }
        if (i != 0)
            env->halted = 1;
        if (smp_cpus > 1) {
            /* XXX: enable it in all cases */
            env->cpuid_features |= CPUID_APIC;
        }
        qemu_register_reset(main_cpu_reset, env);
        if (pci_enabled) {
            apic_init(env);
        }
    }

    /* allocate RAM */
    for (i = 0; i < 512; i++) {
        ram_addr[i] = qemu_ram_alloc(0x8000);
    }
    for (addr = 0; addr < 0xa0000; addr += 0x8000) {
        cpu_register_physical_memory(0x00000000 | addr, 0x8000, ram_addr[addr >> 15]);
        cpu_register_physical_memory(0xfff00000 | addr, 0x8000, ram_addr[addr >> 15]);
    }
    for (addr = 0x100000; addr < 0x1000000; addr += 0x8000) {
        cpu_register_physical_memory(0x00000000 | addr, 0x8000, ram_addr[addr >> 15]);
    }
    if (ram_size > 0xf00000) {
        ram_addr_t ram_addr = qemu_ram_alloc(ram_size - 0xf00000);
        cpu_register_physical_memory(0x1000000, ram_size - 0xf00000, ram_addr);
    }

    /* BIOS load */
    snprintf(buf, sizeof(buf), "%s/%s", bios_dir, BIOS_FILENAME);
    bios_addr = qemu_ram_alloc(BIOS_FILESIZE);
    ret = load_image(buf, qemu_get_ram_ptr(bios_addr));
    if (ret != BIOS_FILESIZE) {
        fprintf(stderr, "qemu: could not load PC98 BIOS '%s'\n", buf);
        exit(1);
    }
    snprintf(buf, sizeof(buf), "%s/%s", bios_dir, ITF_FILENAME);
    itf_addr = qemu_ram_alloc(ITF_FILESIZE);
    ret = load_image(buf, qemu_get_ram_ptr(itf_addr));
    if (ret != ITF_FILESIZE) {
        fprintf(stderr, "qemu: could not load PC98 ITF '%s'\n", buf);
        exit(1);
    }
    snprintf(buf, sizeof(buf), "%s/%s", bios_dir, IDE_FILENAME);
    ide_addr = qemu_ram_alloc(IDE_FILESIZE);
    ret = load_image(buf, qemu_get_ram_ptr(ide_addr));
    if (ret != IDE_FILESIZE) {
        fprintf(stderr, "qemu: could not load PC98 IDE BIOS '%s'\n", buf);
        exit(1);
    }

    cpu_register_physical_memory(0x000d8000, 0x2000, ide_addr | IO_MEM_ROM);
    cpu_register_physical_memory(0x000f8000, 0x8000, itf_addr | IO_MEM_ROM);
    cpu_register_physical_memory(0xfffd8000, 0x2000, ide_addr | IO_MEM_ROM);
    cpu_register_physical_memory(0xffff8000, 0x8000, itf_addr | IO_MEM_ROM);

    bios_ram_selected = 0;
    itf_selected = 1;
    ram_window_map = 0x08;

    cpu_irq = qemu_allocate_irqs(pic_irq_request, NULL, 1);
    i8259 = pc98_i8259_init(cpu_irq[0]);
    ferr_irq = i8259[8];

    if (pci_enabled) {
        pci_bus = i440fx_init(&i440fx_state, i8259);
        piix3_devfn = piix3_init(pci_bus, -1);
    } else {
        pci_bus = NULL;
    }

    /* init basic PC hardware */
    register_ioport_write(0xf0, 1, 1, ioportF0_write, env);
    register_ioport_write(0xf2, 1, 1, ioportF2_write, env);
    register_ioport_read(0xf2, 1, 1, ioportF2_read, env);
    register_ioport_write(0xf6, 1, 1, ioportF6_write, env);
    register_ioport_read(0xf6, 1, 1, ioportF6_read, env);
    register_ioport_write(0x43d, 1, 1, ioport43D_write, env);
    register_ioport_write(0x461, 1, 1, ioport461_write, env);
    register_ioport_read(0x461, 1, 1, ioport461_read, env);
    register_ioport_read(0x534, 1, 1, ioport534_read, env);
    register_ioport_write(0x53d, 1, 1, ioport53D_write, env);
    register_ioport_read(0x63d, 1, 1, ioport63D_read, env);
    register_ioport_read(0x9894, 1, 1, ioport9894_read, env);

    pc98_vga_init(i8259[2]);

    if (pci_enabled) {
        ioapic = ioapic_init();
    }
    pit = pc98_pit_init(i8259[0]);
    /*pcspk_init(pit);*/
    if (pci_enabled) {
        pic_set_alt_irq_func(isa_pic, ioapic_set_irq, ioapic);
    }

/*
    for(i = 0; i < MAX_SERIAL_PORTS; i++) {
        if (serial_hds[i]) {
            serial_init(serial_io[i], i8259[serial_irq[i]], 115200,
                        serial_hds[i]);
        }
    }

    for(i = 0; i < MAX_PARALLEL_PORTS; i++) {
        if (parallel_hds[i]) {
            parallel_init(parallel_io[i], i8259[parallel_irq[i]],
                          parallel_hds[i]);
        }
    }

    for(i = 0; i < nb_nics; i++) {
        NICInfo *nd = &nd_table[i];

        if (!pci_enabled || (nd->model && strcmp(nd->model, "ne2k_isa") == 0))
            pc_init_ne2k_isa(nd, i8259);
        else
            pci_nic_init(pci_bus, nd, -1, "ne2k_pci");
    }

    qemu_system_hot_add_init();
*/

    if (drive_get_max_bus(IF_IDE) >= MAX_IDE_BUS) {
        fprintf(stderr, "qemu: too many IDE bus\n");
        exit(1);
    }

    for(i = 0; i < MAX_IDE_BUS * MAX_IDE_DEVS; i++) {
        index = drive_get_index(IF_IDE, i / MAX_IDE_DEVS, i % MAX_IDE_DEVS);
	if (index != -1)
	    hd[i] = drives_table[index].bdrv;
	else
	    hd[i] = NULL;
    }

    if (pci_enabled) {
        pci_piix3_ide_init(pci_bus, hd, piix3_devfn + 1, i8259);
    } else {
        pc98_ide_init(i8259[9], hd[0], hd[1], hd[2], hd[3]);
    }

    pc98_kbd_init(i8259[1]);
    pc98_mouse_init(i8259[13]);
    pc98_sys_init(i8259[15]);
    pc98_DMA_init(1);
#ifdef HAS_AUDIO
    /*audio_init(pci_enabled ? pci_bus : NULL, i8259);*/
#endif

    for(i = 0; i < MAX_FD; i++) {
        index = drive_get_index(IF_FLOPPY, 0, i);
	if (index != -1)
	    fd[i] = drives_table[index].bdrv;
	else
	    fd[i] = NULL;
    }
    floppy_controller = pc98_fdctrl_init(i8259[11], 2, fd);

    if (pci_enabled && usb_enabled) {
        usb_uhci_piix3_init(pci_bus, piix3_devfn + 2);
    }

    if (i440fx_state) {
        i440fx_init_memory_mappings(i440fx_state);
    }
}

static void pc98_init_pci(ram_addr_t ram_size, int vga_ram_size,
                          const char *boot_device,
                          const char *kernel_filename,
                          const char *kernel_cmdline,
                          const char *initrd_filename,
                          const char *cpu_model)
{
    pc98_init1(ram_size, 1);
}

static void pc98_init_cbus(ram_addr_t ram_size, int vga_ram_size,
                           const char *boot_device,
                           const char *kernel_filename,
                           const char *kernel_cmdline,
                           const char *initrd_filename,
                           const char *cpu_model)
{
    pc98_init1(ram_size, 0);
}

QEMUMachine pc98_machine = {
    .name = "pc98",
    .desc = "NEC PC-98x1",
    .init = pc98_init_cbus,
    .max_cpus = 1,
};

QEMUMachine pc98pci_machine = {
    .name = "pc98pci",
    .desc = "NEC PC-98x1 with PCI",
    .init = pc98_init_pci,
    .max_cpus = 2,
};
